//! Bundle all functions together.

use std::{
    io,
    path::{Path, PathBuf},
};

use r2d2::Pool;
use r2d2_sqlite::SqliteConnectionManager;
use snafu::{ResultExt, Snafu};
use tempfile::{NamedTempFile, PersistError};

use crate::{
    data::{self, author::Author, book::Book, pagination::SortOrder, series::Series},
    search::search,
};

/// Top level calibre functions, bundling all sub functions in one place and providing secure access to
/// the database.
#[derive(Debug, Clone)]
pub struct Calibre {
    pool: Pool<SqliteConnectionManager>,
    search_db_path: PathBuf,
}

#[derive(Debug, Snafu)]
/// Errors that can occur when loading the calibre database.
pub enum LoadError {
    /// A failure to create the database connection pool.
    #[snafu(display("Failed to create database connection pool."))]
    CreateDbPool { source: r2d2::Error },
    /// A failure to create the temporary database view.
    #[snafu(display("Failed to create temporary database view."))]
    TmpDb { source: io::Error },
    /// A failure to keep the temporary database from being deleted.
    #[snafu(display("Failed to keep temporary database from deletion."))]
    PersistTmpDb { source: PersistError },
}

#[derive(Debug, Snafu)]
/// Errors that can occur when accessing the data store.
pub enum DataStoreError {
    /// A failure to get a database connection from the pool.
    #[snafu(display("Failed to get database connection from pool."))]
    GetDbConn { source: r2d2::Error },

    /// A failure during a search operation.
    #[snafu(display("Failed to search."))]
    Search { source: crate::search::SearchError },

    /// A failure to fetch multiple books.
    #[snafu(display("Failed to fetch multiple books."))]
    MultipleBooks {
        source: data::book::MultipleBooksError,
    },
    /// A failure to fetch recent books.
    #[snafu(display("Failed to fetch recent books."))]
    RecentBooks {
        source: data::book::RecentBooksError,
    },
    /// A failure to fetch a single book.
    #[snafu(display("Failed to fetch book."))]
    ScalarBook { source: data::book::ScalarBookError },
    /// A failure to get a series' books.
    #[snafu(display("Failed to get a series' books."))]
    SeriesBooks { source: data::book::SeriesBookError },
    /// A failure to fetch an author's books.
    #[snafu(display("Failed to fetch an author's books."))]
    AuthorBooks {
        source: data::book::AuthorBooksError,
    },
    /// A failure to check for previous books.
    #[snafu(display("Failed to check if there are previous books."))]
    HasPreviousBooks {
        source: data::book::PreviousBooksError,
    },
    /// A failure to check for more books.
    #[snafu(display("Failed to check if there are more books."))]
    HasMoreBooks { source: data::book::MoreBooksError },

    /// A failure to fetch multiple authors.
    #[snafu(display("Failed to fetch multiple authors."))]
    MultipleAuthors {
        source: data::author::MultipleAuthorsError,
    },
    /// A failure to fetch a single author.
    #[snafu(display("Failed to fetch author."))]
    ScalarAuthor {
        source: data::author::ScalarAuthorError,
    },
    /// A failure to fetch a book's author.
    #[snafu(display("Failed to fetch book's author."))]
    BookAuthor {
        source: data::author::BookAuthorError,
    },
    /// A failure to check for previous authors.
    #[snafu(display("Failed to check if there are previous authors."))]
    HasPreviousAuthors {
        source: data::author::PreviousAuthorsError,
    },
    /// A failure to check for more authors.
    #[snafu(display("Failed to check if there are more authors."))]
    HasMoreAuthors {
        source: data::author::MoreAuthorsError,
    },

    /// A failure to fetch multiple series.
    #[snafu(display("Failed to fetch multiple series."))]
    MultipleSeries {
        source: data::series::MultiplSeriesError,
    },
    /// A failure to fetch a single series.
    #[snafu(display("Failed to fetch series."))]
    ScalarSeries {
        source: data::series::ScalarSeriesError,
    },
    /// A failure to get the series a book belongs to.
    #[snafu(display("Failed to get the series a book belongs to."))]
    BookSeries {
        source: data::series::SeriesBooksError,
    },
    /// A failure to check for previous series.
    #[snafu(display("Failed to check if there are previous series."))]
    HasPreviousSeries {
        source: data::series::PreviousSeriesError,
    },
    /// A failure to check for more series.
    #[snafu(display("Failed to check if there are more series."))]
    HasMoreSeries {
        source: data::series::MoreSeriesError,
    },
}
impl Calibre {
    /// Open a connection to the calibre database.
    ///
    /// Fail if the database file can not be opened or not be found.
    pub fn load(path: &Path) -> Result<Self, LoadError> {
        let manager = SqliteConnectionManager::file(path);
        let pool = r2d2::Pool::new(manager).context(CreateDbPoolSnafu)?;

        let tmpfile = NamedTempFile::new().context(TmpDbSnafu)?;
        let (_, search_db_path) = tmpfile.keep().context(PersistTmpDbSnafu)?;

        Ok(Self {
            pool,
            search_db_path,
        })
    }

    /// Full text search with a query.
    ///
    /// See https://www.sqlite.org/fts5.html#full_text_query_syntax for syntax.
    pub fn search(&self, query: &str) -> Result<Vec<Book>, DataStoreError> {
        search(query, &self.pool, &self.search_db_path).context(SearchSnafu)
    }

    /// Fetch book data from calibre, starting at `cursor`, fetching up to an amount of `limit` and
    /// ordering by `sort_order`.
    pub fn books(
        &self,
        limit: u64,
        cursor: Option<&str>,
        sort_order: &SortOrder,
    ) -> Result<Vec<Book>, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Book::multiple(&conn, limit, cursor, sort_order).context(MultipleBooksSnafu)
    }

    /// Fetch author data from calibre, starting at `cursor`, fetching up to an amount of `limit` and
    /// ordering by `sort_order`.
    pub fn authors(
        &self,
        limit: u64,
        cursor: Option<&str>,
        sort_order: &SortOrder,
    ) -> Result<Vec<Author>, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Author::multiple(&conn, limit, cursor, sort_order).context(MultipleAuthorsSnafu)
    }

    /// Fetch books for an author specified by `author_id`, paginate the books by starting at `cursor`,
    /// fetching up to an amount of `limit` and ordering by `sort_order`.
    pub fn author_books(
        &self,
        author_id: u64,
        limit: u64,
        cursor: Option<&str>,
        sort_order: SortOrder,
    ) -> Result<Vec<Book>, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Book::author_books(&conn, author_id, limit, cursor, sort_order).context(AuthorBooksSnafu)
    }

    /// Get recent books up to a limit of `limit`.
    pub fn recent_books(&self, limit: u64) -> Result<Vec<Book>, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Book::recents(&conn, limit).context(RecentBooksSnafu)
    }

    /// Get a single book, specified `id`.
    pub fn scalar_book(&self, id: u64) -> Result<Book, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Book::scalar_book(&conn, id).context(ScalarBookSnafu)
    }

    /// Get the author to a book with id `id`.
    pub fn book_author(&self, id: u64) -> Result<Author, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Author::book_author(&conn, id).context(BookAuthorSnafu)
    }

    /// Fetch series data from calibre, starting at `cursor`, fetching up to an amount of `limit` and
    /// ordering by `sort_order`.
    pub fn series(
        &self,
        limit: u64,
        cursor: Option<&str>,
        sort_order: &SortOrder,
    ) -> Result<Vec<Series>, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Series::multiple(&conn, limit, cursor, sort_order).context(MultipleSeriesSnafu)
    }

    /// Get the series a book with id `id` is in, as well as the book's position within the series.
    pub fn book_series(&self, id: u64) -> Result<Option<(Series, f64)>, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Series::book_series(&conn, id).context(BookSeriesSnafu)
    }

    /// Get all books belonging to the series with id `id`.
    pub fn series_books(&self, id: u64) -> Result<Vec<Book>, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Book::series_books(&conn, id).context(SeriesBooksSnafu)
    }

    /// Check if there are more authors before the specified cursor.
    pub fn has_previous_authors(&self, author_sort: &str) -> Result<bool, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Author::has_previous_authors(&conn, author_sort).context(HasPreviousAuthorsSnafu)
    }

    /// Check if there are more authors after the specified cursor.
    pub fn has_more_authors(&self, author_sort: &str) -> Result<bool, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Author::has_more_authors(&conn, author_sort).context(HasMoreAuthorsSnafu)
    }

    /// Check if there are more books before the specified cursor.
    pub fn has_previous_books(&self, book_sort: &str) -> Result<bool, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Book::has_previous_books(&conn, book_sort).context(HasPreviousBooksSnafu)
    }

    /// Check if there are more books after the specified cursor.
    pub fn has_more_books(&self, book_sort: &str) -> Result<bool, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Book::has_more_books(&conn, book_sort).context(HasMoreBooksSnafu)
    }

    /// Check if there are more series before the specified cursor.
    pub fn has_previous_series(&self, series_sort: &str) -> Result<bool, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Series::has_previous_series(&conn, series_sort).context(HasPreviousSeriesSnafu)
    }

    /// Check if there are more series after the specified cursor.
    pub fn has_more_series(&self, series_sort: &str) -> Result<bool, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Series::has_more_series(&conn, series_sort).context(HasMoreSeriesSnafu)
    }

    /// Fetch a single author with id `id`.
    pub fn scalar_author(&self, id: u64) -> Result<Author, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Author::scalar_author(&conn, id).context(ScalarAuthorSnafu)
    }

    /// Fetch a single series with id `id`.
    pub fn scalar_series(&self, id: u64) -> Result<Series, DataStoreError> {
        let conn = self.pool.get().context(GetDbConnSnafu)?;
        Series::scalar_series(&conn, id).context(ScalarSeriesSnafu)
    }
}

#[cfg(test)]
/// Tests for the calibre module.
mod tests {
    use super::*;

    fn init_calibre() -> Calibre {
        Calibre::load(Path::new("./testdata/metadata.db")).unwrap()
    }

    #[test]
    fn books() {
        let c = init_calibre();
        let books = c.books(10, None, &SortOrder::ASC).unwrap();
        assert_eq!(books.len(), 4);
    }

    #[test]
    fn authors() {
        let c = init_calibre();
        let authors = c.authors(10, None, &SortOrder::ASC).unwrap();
        assert_eq!(authors.len(), 3);
    }

    #[test]
    fn author_books() {
        let c = init_calibre();
        let books = c.author_books(1, 10, None, SortOrder::ASC).unwrap();
        assert_eq!(books.len(), 2);
    }

    #[test]
    fn pagination() {
        let c = init_calibre();
        let authors = c.authors(1, None, &SortOrder::ASC).unwrap();
        assert_eq!(authors.len(), 1);
        assert_eq!(authors[0].name, "Kevin R. Grazier");

        let authors = c
            .authors(1, Some(&authors[0].sort), &SortOrder::ASC)
            .unwrap();
        assert_eq!(authors.len(), 1);
        assert_eq!(authors[0].name, "Terry Pratchett");

        let authors = c
            .authors(1, Some(&authors[0].sort), &SortOrder::ASC)
            .unwrap();
        assert_eq!(authors.len(), 1);
        assert_eq!(authors[0].name, "Edward Noyes Westcott");

        let authors = c
            .authors(1, Some(&authors[0].sort), &SortOrder::DESC)
            .unwrap();
        assert_eq!(authors.len(), 1);
        assert_eq!(authors[0].name, "Terry Pratchett");
    }
}
